/***************************************************************************
                          MovingMesh.cpp  -  description
                             -------------------
    begin                : Wed Jun 20 2001
    copyright            : (C) 2001 by Robert Lie
    email                : rli@math.hkbu.edu.hk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "MovingMesh.h"



LRTriangulation::LRTriangulation ()
{
}



LRTriangulation::~LRTriangulation ()
{
}



void
LRTriangulation::ReadEasymesh (const string filename)
{
	unsigned int i, j, k;
	char buffer[256];
	ifstream is;

	mesh_data_file = filename;

	is.open ((mesh_data_file + ".n").c_str ());
	std::cout << "Reading nodes data ..." << std::endl;
	is >> n_node >> n_element >> n_face;
	node.resize (n_node);
	face.resize (n_face);
	element.resize (n_element);
	is.getline (buffer, 256);
	n_interior_node = 0;
	n_boundary_node = 0;
	for (i = 0; i < n_node; i++) {
		is >> j >> node[i].coordinate[0]
			>> node[i].coordinate[1] >> node[i].boundary_mark;
		if (node[i].boundary_mark)
			node[i].index = n_boundary_node++;
		else
			node[i].index = n_interior_node++;
	}
	is.close ();

	is.open ((mesh_data_file + ".s").c_str ());
	std::cout << "Reading faces data ..." << std::endl;
	is >> j;
	assert (j == n_face);
	for (i = 0; i < n_face; i++) {
		is >> j >> face[i].node[0]
			>> face[i].node[1]
			>> face[i].neighbour_element[0]
			>> face[i].neighbour_element[1]
			>> face[i].boundary_mark;
	}
	is.close ();

	is.open ((mesh_data_file + ".e").c_str ());
	std::cout << "Reading elements data ..." << std::endl;
	is >> i >> j >> k;
	assert (j == n_node);
	assert (k == n_face);
	assert (i == n_element);
	is.getline (buffer, 256);
	for (i = 0; i < n_element; i++) {
		is >> j >> element[i].node[0]
			>> element[i].node[1]
			>> element[i].node[2]
			>> element[i].neighbour_element[0]
			>> element[i].neighbour_element[1]
			>> element[i].neighbour_element[2]
			>> element[i].face[0]
			>> element[i].face[1] >> element[i].face[2];
	}
	is.close ();

	ParseMesh ();
}



void
LRTriangulation::ParseMesh ()
{
	unsigned int i, j, k, l, m;
	interior_node_index.resize (n_interior_node);
	boundary_node_index.resize (n_boundary_node);
	j = 0;
	k = 0;
	for (i = 0; i < n_node; i++) {
		if (node[i].boundary_mark)
			boundary_node_index[k++] = i;
		else
			interior_node_index[j++] = i;
	}
	//----generate the index of variations------------------------
	index.resize (n_node, -1);
	for (i = 0, n_new_node = 0;i < n_node;i ++) {
		if (node[i].boundary_mark == 0) { // this is an interior node
			index[i] = n_new_node ++;
			continue;
		}
		if (index[i] != -1) continue;
		for (k = 0;k < n_boundary_node;k ++) {
			l = boundary_node_index[k];
			if (index[l] == -1) continue;
			double dx = fabs(node[i].coordinate[0] - node[l].coordinate[0]);
			double dy = fabs(node[i].coordinate[1] - node[l].coordinate[1]);
			if (dx*(1.0 - dx) <= 1.0e-3 && dy*(1.0 - dy) <= 1.0e-3) {
				index[i] = index[l];
				break;
			}
		}
		if (k == n_boundary_node) index[i] = n_new_node ++;
	}
	std::cout << "# dof = " << n_new_node << std::endl;
	
	//find the periodic neibour element of the face
	std::list<int> bound_face;
	for (i = 0;i < n_face;i ++)
		if (face[i].boundary_mark != 0)
			bound_face.push_back(i);
	do {
		i = bound_face.front();
		bound_face.pop_front();
		std::list<int>::iterator the_int = bound_face.begin();
		std::list<int>::iterator end_int = bound_face.end();
		int& ln0 = face[i].neighbour_element[0];
		int& rn0 = face[i].neighbour_element[1];
		for (;the_int != end_int;the_int ++) {
			j = *the_int;
			if ((index[face[i].node[0]] == index[face[j].node[0]]
				&& index[face[i].node[1]] == index[face[j].node[1]])
				|| (index[face[i].node[0]] == index[face[j].node[1]]
				&& index[face[i].node[1]] == index[face[j].node[0]])) {
				int& ln1 = face[j].neighbour_element[0];
				int& rn1 = face[j].neighbour_element[1];
				if (ln0 == -1 && rn0 >= 0) {
					if (ln1 == -1 && rn1 >= 0) {
						ln0 = rn1;
						ln1 = rn0;
					}
					else if (ln1 >= 0 && rn1 == -1) {
						ln0 = ln1;
						rn1 = rn0;
					}
					else assert(false);
				}
				else if(ln0 >= 0 && rn0 == -1) {
					if (ln1 == -1 && rn1 >= 0) {
						rn0 = rn1;
						ln1 = ln0;
					}
					else if (ln1 >= 0 && rn1 == -1) {
						rn0 = ln1;
						rn1 = ln0;
					}
					else assert(false);
				}
				else assert(false);
				bound_face.erase(the_int);
				break;
			}
		}
		assert(the_int != end_int);
	} while (!bound_face.empty());
}



LRMovingMesh::LRMovingMesh ()
{
}



LRMovingMesh::~LRMovingMesh ()
{
}



void
LRMovingMesh::ReadDomain (const string filename)
{
	LRTriangulation::ReadEasymesh (filename);
	logical_node.resize (n_node);
	move_direction.resize (n_node);
	logical_move_direction.resize (n_node);
	monitor_function.resize (n_element);

	unsigned int i, j, k, l, m, l0, l1, special_node_index;
	double d;
	ifstream is;

	is.open ((mesh_data_file + ".d").c_str ());
	ReadDummy (is);
	is >> domain.n_vertex;
	domain.physical_domain_vertex.resize (domain.n_vertex);
	domain.logical_domain_vertex.resize (domain.n_vertex);
	for (i = 0; i < domain.n_vertex; i++)
	{
		ReadDummy (is);
		is >> domain.physical_domain_vertex[i].index;
		ReadDummy (is);
		is >> domain.physical_domain_vertex[i].coordinate[0];
		ReadDummy (is);
		is >> domain.physical_domain_vertex[i].coordinate[1];
		ReadDummy (is);
		is >> d;
		ReadDummy (is);
		is >> domain.physical_domain_vertex[i].boundary_mark;
	}
	ReadDummy (is);
	is >> domain.n_edge;
	domain.edge.resize (domain.n_edge);
	for (i = 0; i < domain.n_edge; i++)
	{
		ReadDummy (is);
		is >> domain.edge[i].index;
		ReadDummy (is);
		is >> domain.edge[i].vertex[0];
		ReadDummy (is);
		is >> domain.edge[i].vertex[1];
		ReadDummy (is);
		is >> domain.edge[i].boundary_mark;
	}
	is.close ();
	is.open ((mesh_data_file + ".log").c_str ());
	for (i = 0; i < domain.n_vertex; i++)
	{
		is >> domain.logical_domain_vertex[i].coordinate[0];
		is >> domain.logical_domain_vertex[i].coordinate[1];
	}
	is.close ();

	std::vector < unsigned int >n_coupling_node (n_new_node, 0);
	for (i = 0; i < n_face; i++) {
		n_coupling_node[index[face[i].node[0]]]++;
		n_coupling_node[index[face[i].node[1]]]++;
	}
	n_max_coupling_node = 0;
	for (i = 0; i < n_new_node; i++)
		n_max_coupling_node = max (n_max_coupling_node, n_coupling_node[i]);
	n_max_coupling_node++;
	spM.reinit (n_new_node - 1, n_new_node - 1, n_max_coupling_node);
	for (special_node_index = 0;special_node_index < n_node;special_node_index ++)
		if (node[special_node_index].boundary_mark == 0) break;
	special_node_index = index[special_node_index];
	for (i = 0; i < n_element; i++) {
		for (j = 0; j < 3; j++) {
			l0 = index[element[i].node[j]];
			if (l0 == special_node_index) continue;
			if (l0 > special_node_index) l0 --;
			for (k = 0; k < 3; k++) {
				l1 = index[element[i].node[k]];
				if (l1 == special_node_index) continue;
				if (l1 > special_node_index) l1 --;
				spM.add (l0, l1);
			}
		}
	}
	spM.compress ();

	std::vector < unsigned int >n_couple_N (n_new_node * 3, 0);
	for (i = 0; i < n_element; i++) {
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			n_couple_N[index[k]] += 6;
			n_couple_N[n_new_node + index[k]] += 6;
			n_couple_N[n_new_node * 2 + index[k]] += 6;
		}
	}
	n_max_coupling_node = 0;
	for (i = 0; i < n_new_node * 3; i++)
		n_max_coupling_node = max (n_max_coupling_node, n_couple_N[i]);

//needn't avoid boundary nodes
	spM_n.reinit (3 * n_new_node - 1, 3 * n_new_node - 1, n_max_coupling_node);
	for (i = 0; i < n_element; i++) {
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			for (l = 0; l < 3; l++) {
				m = element[i].node[l];
				if (index[m] != n_new_node - 1) {
					spM_n.add (index[k], n_new_node * 2 + index[m]);
					spM_n.add (n_new_node * 2 + index[m], index[k]);
					spM_n.add (n_new_node + index[k], n_new_node * 2 + index[m]);
					spM_n.add (n_new_node * 2 + index[m], n_new_node + index[k]);
				}
				spM_n.add (index[k], index[m]);
				spM_n.add (n_new_node + index[k], n_new_node + index[m]);
			}
		}
	}
	spM_n.compress ();

	std::cout << "Computing logical mesh ..." << std::endl;
	GetLogicalMesh ();
}



void
LRMovingMesh::MoveMesh ()
{
	unsigned i, j;
	double epsilon = 0.01;
	double error = 1.;
	while (error > epsilon) {
		GetMoveDirection ();
		GetMoveStepLength ();
		UpdateSolution ();
		for (i = 0; i < n_node; i++) {
			node[i].coordinate[0] += move_step_length * move_direction[i].coordinate[0];
			node[i].coordinate[1] += move_step_length * move_direction[i].coordinate[1];
		}
		error = 0;
		for (i = 0; i < n_node; i++) {
			for (j = 0; j < 2; j++)	{
				double d = fabs (move_direction[i].coordinate[j]);
				error = max (error, d * move_step_length);
			}
		}
		std::cout << "\r\terror = " << error << "\n" << std::flush;
	}
}



void
LRMovingMesh::OutputPhysicalMesh (const string filename)
{
	unsigned int i;
	ofstream is (filename.c_str ());
	is << "VARIABLES=\"X\",\"Y\"\n ";
	is << "ZONE N=" << n_node << ", E=" << n_element <<
		", F=FEPOINT, ET=TRIANGLE\n";
	for (i = 0; i < n_node; i++)
	{
		is << node[i].coordinate[0] << "   ";
		is << node[i].coordinate[1] << "   ";
		is << "\n";
	}
	is << "\n";
	for (i = 0; i < n_element; i++)
	{
		is << element[i].node[0] + 1 << "  "
			<< element[i].node[1] + 1 << "  "
			<< element[i].node[2] + 1 << "  " << "\n";
	}
	is.close ();
}



void
LRMovingMesh::OutputLogicalMesh (const string filename)
{
	unsigned int i;
	ofstream is ((filename + ".n").c_str ());
	is << "VARIABLES=\"X\",\"Y\"\n ";
	is << "ZONE N=" << n_node << ", E=" << n_element <<
		", F=FEPOINT, ET=TRIANGLE\n";
	for (i = 0; i < n_node; i++)
	{
		is << logical_node[i].coordinate[0] << '\t'
			<< logical_node[i].coordinate[1] << '\t';
		is << "\n";
	}
	is << "\n";
	for (i = 0; i < n_element; i++)
	{
		is << element[i].node[0] + 1 << "  "
			<< element[i].node[1] + 1 << "  "
			<< element[i].node[2] + 1 << "  " << "\n";
	}
	is.close ();
}



void
LRMovingMesh::GetMoveDirection ()
{
	unsigned int i, j, k, l0, l1;
	int special_node, special_node_index;
	for (special_node = 0;special_node < n_node;special_node ++)
		if (node[special_node].boundary_mark == 0) break;
	special_node_index = index[special_node];

	MonitorFunction ();
	SparseMatrix<double> M(spM);
	Vector<double> r0(n_new_node - 1);
	Vector<double> r1(n_new_node - 1);
	for (i = 0; i < n_element; i++) {
		double omega[2][3];
		double area = (node[element[i].node[1]].coordinate[0] - node[element[i].node[0]].coordinate[0]) *
			(node[element[i].node[2]].coordinate[1] - node[element[i].node[0]].coordinate[1]) -
			(node[element[i].node[1]].coordinate[1] - node[element[i].node[0]].coordinate[1]) *
			(node[element[i].node[2]].coordinate[0] - node[element[i].node[0]].coordinate[0]);
		omega[0][0] = node[element[i].node[2]].coordinate[0] - node[element[i].node[1]].coordinate[0];
		omega[0][1] = node[element[i].node[0]].coordinate[0] - node[element[i].node[2]].coordinate[0];
		omega[0][2] = node[element[i].node[1]].coordinate[0] - node[element[i].node[0]].coordinate[0];
		omega[1][0] = node[element[i].node[2]].coordinate[1] - node[element[i].node[1]].coordinate[1];
		omega[1][1] = node[element[i].node[0]].coordinate[1] - node[element[i].node[2]].coordinate[1];
		omega[1][2] = node[element[i].node[1]].coordinate[1] - node[element[i].node[0]].coordinate[1];
		for (j = 0; j < 3; j++) {
			l0 = index[element[i].node[j]];
			if (l0 == special_node_index) continue;
			if (l0 > special_node_index) l0 --;
			for (k = 0; k < 3; k++) {
				double d = monitor_function[i]*(omega[0][j]*omega[0][k] + omega[1][j]*omega[1][k])/area;
				l1 = index[element[i].node[k]];
				if (l1 == special_node_index) {
					r0(l0) -= d*(logical_node[element[i].node[k]].coordinate[0] - node[element[i].node[k]].coordinate[0]);
					r1(l0) -= d*(logical_node[element[i].node[k]].coordinate[1] - node[element[i].node[k]].coordinate[1]);
					continue;
				}
				if (l1 > special_node_index) l1 --;
				M.add (l0, l1, d);
			}
			r0(l0) += monitor_function[i]*omega[1][j];
			r1(l0) -= monitor_function[i]*omega[0][j];
		}
	}
	Vector<double> x(n_new_node - 1);
	SparseILU<double> ilu (spM);
	ilu.decompose(M);
	SolverControl solver_control (200, 1e-6);
	PrimitiveVectorMemory <> vector_memory;
	SolverBicgstab<> bicgstab (solver_control, vector_memory);
	bicgstab.solve (M, x, r0,
		PreconditionUseMatrix <SparseILU<double>, Vector<double> >
			(ilu, &SparseILU<double>::template apply_decomposition<double>));
	for (i = 0; i < n_node; i ++) {
		if (index[i] == special_node_index)
			logical_move_direction[i].coordinate[0] = 0.0;
		else if (index[i] < special_node_index)
			logical_move_direction[i].coordinate[0] = logical_node[i].coordinate[0] 
				- (x(index[i]) + node[i].coordinate[0]);
		else
			logical_move_direction[i].coordinate[0] = logical_node[i].coordinate[0] 
				- (x(index[i] - 1) + node[i].coordinate[0]);
	}

	bicgstab.solve (M, x, r1,
		PreconditionUseMatrix <SparseILU<double>, Vector<double> >
			(ilu, &SparseILU<double>::template apply_decomposition<double>));
	for (i = 0; i < n_node; i++) {
		if (index[i] == special_node_index)
			logical_move_direction[i].coordinate[1] = 0.0;
		else if (index[i] < special_node_index)
			logical_move_direction[i].coordinate[1] = logical_node[i].coordinate[1] 
				- (x(index[i]) + node[i].coordinate[1]);
		else
			logical_move_direction[i].coordinate[1] = logical_node[i].coordinate[1]
				- (x(index[i] - 1) + node[i].coordinate[1]);
	}
	
	std::vector<LRPoint> temp_move_direction(n_new_node);
	for (i = 0; i < n_new_node; i++) {
		temp_move_direction[i].coordinate[0] = 0;
		temp_move_direction[i].coordinate[1] = 0;
	}
	std::vector<double> mass_accumulation (n_new_node, 0.);
	for (i = 0; i < n_element; i++) {
		double area;
		double jacobi[2][2];
		area = (logical_node[element[i].node[1]].coordinate[0] - logical_node[element[i].node[0]].coordinate[0]) *
			(logical_node[element[i].node[2]].coordinate[1] - logical_node[element[i].node[0]].coordinate[1]) -
			(logical_node[element[i].node[1]].coordinate[1] - logical_node[element[i].node[0]].coordinate[1]) *
			(logical_node[element[i].node[2]].coordinate[0] - logical_node[element[i].node[0]].coordinate[0]);
		jacobi[0][0] = (node[element[i].node[1]].coordinate[0] - node[element[i].node[0]].coordinate[0]) *
			(logical_node[element[i].node[2]].coordinate[1] - logical_node[element[i].node[0]].coordinate[1]) -
			(logical_node[element[i].node[1]].coordinate[1] - logical_node[element[i].node[0]].coordinate[1]) *
			(node[element[i].node[2]].coordinate[0] - node[element[i].node[0]].coordinate[0]);
		jacobi[1][0] = (node[element[i].node[1]].coordinate[1] - node[element[i].node[0]].coordinate[1]) *
			(logical_node[element[i].node[2]].coordinate[1] - logical_node[element[i].node[0]].coordinate[1]) -
			(logical_node[element[i].node[1]].coordinate[1] - logical_node[element[i].node[0]].coordinate[1]) *
			(node[element[i].node[2]].coordinate[1] - node[element[i].node[0]].coordinate[1]);
		jacobi[0][1] = (logical_node[element[i].node[1]].coordinate[0] - logical_node[element[i].node[0]].coordinate[0]) *
			(node[element[i].node[2]].coordinate[0] - node[element[i].node[0]].coordinate[0]) -
			(node[element[i].node[1]].coordinate[0] - node[element[i].node[0]].coordinate[0]) *
			(logical_node[element[i].node[2]].coordinate[0] - logical_node[element[i].node[0]].coordinate[0]);
		jacobi[1][1] = (logical_node[element[i].node[1]].coordinate[0] - logical_node[element[i].node[0]].coordinate[0]) *
			(node[element[i].node[2]].coordinate[1] - node[element[i].node[0]].coordinate[1]) -
			(node[element[i].node[1]].coordinate[1] - node[element[i].node[0]].coordinate[1]) *
			(logical_node[element[i].node[2]].coordinate[0] - logical_node[element[i].node[0]].coordinate[0]);
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			temp_move_direction[index[k]].coordinate[0] += jacobi[0][0] * logical_move_direction[k].coordinate[0] +
				jacobi[0][1] * logical_move_direction[k].coordinate[1];
			temp_move_direction[index[k]].coordinate[1] += jacobi[1][0] * logical_move_direction[k].coordinate[0] +
				jacobi[1][1] * logical_move_direction[k].coordinate[1];
			mass_accumulation[index[k]] += area;
		}
	}
	for (i = 0; i < n_node; i++) {
		j = index[i];
		move_direction[i].coordinate[0] = temp_move_direction[j].coordinate[0] / mass_accumulation[j];
		move_direction[i].coordinate[1] = temp_move_direction[j].coordinate[1] / mass_accumulation[j];
	}
	for (i = 0; i < n_node; i++) {
		move_direction[i].coordinate[0] -= move_direction[special_node].coordinate[0];
		move_direction[i].coordinate[1] -= move_direction[special_node].coordinate[1];
	}
}



void
LRMovingMesh::GetMoveStepLength ()
{
	unsigned int i, j, k;

	double a, b, c, max_move;
    move_step_length = 1;
    for (i = 0;i < n_element;i ++) {
            a = (move_direction[element[i].node[1]].coordinate[0] - move_direction[element[i].node[0]].coordinate[0])
                    * (move_direction[element[i].node[2]].coordinate[1] - move_direction[element[i].node[0]].coordinate[1])
                    - (move_direction[element[i].node[1]].coordinate[1] - move_direction[element[i].node[0]].coordinate[1])
                    * (move_direction[element[i].node[2]].coordinate[0] - move_direction[element[i].node[0]].coordinate[0]);
            b = move_direction[element[i].node[0]].coordinate[0] * (node[element[i].node[1]].coordinate[1] - node[element[i].node[2]].coordinate[1])
                    - move_direction[element[i].node[0]].coordinate[1] * (node[element[i].node[1]].coordinate[0] - node[element[i].node[2]].coordinate[0])
                    + move_direction[element[i].node[1]].coordinate[0] * (node[element[i].node[2]].coordinate[1] - node[element[i].node[0]].coordinate[1])
                    - move_direction[element[i].node[1]].coordinate[1] * (node[element[i].node[2]].coordinate[0] - node[element[i].node[0]].coordinate[0])
                    + move_direction[element[i].node[2]].coordinate[0] * (node[element[i].node[0]].coordinate[1] - node[element[i].node[1]].coordinate[1])
                    - move_direction[element[i].node[2]].coordinate[1] * (node[element[i].node[0]].coordinate[0] - node[element[i].node[1]].coordinate[0]);
            c = (node[element[i].node[1]].coordinate[0] - node[element[i].node[0]].coordinate[0])
                    * (node[element[i].node[2]].coordinate[1] - node[element[i].node[0]].coordinate[1])
                    - (node[element[i].node[1]].coordinate[1] - node[element[i].node[0]].coordinate[1])
                    * (node[element[i].node[2]].coordinate[0] - node[element[i].node[0]].coordinate[0]);
            if (a == 0) {
                    if (b == 0)
                            ; //do nothing
                    else if (c/b > 0)
                            ; //do nothing
                    else
                            move_step_length = min(move_step_length, -c/b);
            }
            else if (b*b - 4*a*c < 0)
                    ; //do nothing
            else {
                    if (a < 0) {
                            a = -a;
                            b = -b;
                            c = -c;
                    }
                    double d = (-b - sqrt(b*b - 4*a*c))/(2*a);
                    if (d < 0) {
                            d = (-b + sqrt(b*b - 4*a*c))/(2*a);
                            if (d > 0)
                                    move_step_length = min(move_step_length, d);
                    }
                    else
                            move_step_length = min(move_step_length, d);
            }
    }
    move_step_length *= 0.5;
}



void
LRMovingMesh::ReadDummy (ifstream & is)
{
	char c;
	for (is.get (c); true; is.get (c))
	{
		if (c == '#')
		{
			while (true)
			{
				is.get (c);
				if (c == '#')
					break;
			}
		}
		else if (('0' <= c && c <= '9') || (c == '.') || (c == '-'))
			break;
	}
	is.putback (c);
}



void
LRMovingMesh::MonitorFunction ()
{
}



void
LRMovingMesh::SmoothMonitor (unsigned int s)
{
	unsigned int i, j;
	std::vector < float >area (n_element);
	std::vector < float >mass_accumulation (n_new_node, 0);
	std::vector < float >monitor (n_new_node);
	for (i = 0; i < n_element; i++)	{
		area[i] = (node[element[i].node[1]].coordinate[0] - node[element[i].node[0]].coordinate[0]) *
			(node[element[i].node[2]].coordinate[1] - node[element[i].node[0]].coordinate[1]) -
			(node[element[i].node[1]].coordinate[1] - node[element[i].node[0]].coordinate[1]) *
			(node[element[i].node[2]].coordinate[0] - node[element[i].node[0]].coordinate[0]);
		mass_accumulation[index[element[i].node[0]]] += area[i];
		mass_accumulation[index[element[i].node[1]]] += area[i];
		mass_accumulation[index[element[i].node[2]]] += area[i];
	}
	for (i = 0; i < s; i++)	{
		for (j = 0; j < n_new_node; j++)
			monitor[j] = 0;
		for (j = 0; j < n_element; j++)	{
			monitor[index[element[j].node[0]]] += monitor_function[j] * area[j];
			monitor[index[element[j].node[1]]] += monitor_function[j] * area[j];
			monitor[index[element[j].node[2]]] += monitor_function[j] * area[j];
		}
		for (j = 0; j < n_new_node; j++)
			monitor[j] /= mass_accumulation[j];
		for (j = 0; j < n_element; j++) {
			monitor_function[j] = (1. / 3.) * (monitor[index[element[j].node[0]]] +
				monitor[index[element[j].node[1]]] + monitor[index[element[j].node[2]]]);
		}
	}
}



void
LRMovingMesh::UpdateSolution ()
{
}



void
LRMovingMesh::GetLogicalMesh ()
{
	unsigned int i;
	for (i = 0;i < n_node;i ++) {
		logical_node[i].coordinate[0] = node[i].coordinate[0];
		logical_node[i].coordinate[1] = node[i].coordinate[1];
	}
}


/*
 * end of file
 ***************************************************************************/
