/***************************************************************************
                          stabledls.cpp  -  description
                             -------------------
    begin                :  10 30 2002
    copyright            : (C) 2002 by 
    email                : dyn@DYN
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "stabledls.h"
#define PI			(4.0*atan(1.0))

#define ILU_BICG_SOLVER

movedls::movedls(const string& data_path) :
	DATA_PATH("/home/dyn/result_dsl/" + data_path + "/")
{
}

movedls::~movedls()
{
}

void movedls::InitialValue()
{
	unsigned int i;
	double x, y, ro= 100., delta= 0.05;
	Nu=1./20000.;
	for(i=0;i<n_node;i++){
		x = node[i].coordinate[0];
		y = node[i].coordinate[1];
		if(y <= 0.5)    solution(i)= tanh(ro*(y-0.25));
		else            solution(i)= tanh(ro*(0.75-y));
		solution(i+n_node)= delta*sin(2.*PI*x);
		if(i != n_node-1)
			 solution(i+2*n_node)= 0;
	}
}

/**
 * calculate the matrix $1/dt + nu \Delta$ and $\nabla$ on an element
 */
void movedls::ele_matrix_AB (double node[3][2], double dt, double nu,
	 double A[3][3], double B[6])
{
	int i, j;
	double M[3][3],	u0[6], g[3][2];
	func_on_ele u (u0, node);
	double S = fabs (u.area) / 2.;
	for (i = 0; i < 3; i++)	{
		g[i][0] = u.ita[i] / u.area;
		g[i][1] = -u.ata[i] / u.area;
	}

	for (i = 0; i < 3; i++) {
		M[i][i] = 1./6.;
		for (j = i + 1; j < 3; j++) {
			M[i][j] = M[j][i] = 1./12.;
		}
	}
	for (i = 0; i < 3; i++) {
		for (j = 0; j < 3; j++) {
			A[i][j] = g[i][0]*g[j][0] + g[i][1]*g[j][1];
		}
	}
	for (i = 0; i < 3; i++) {
		for (j = 0; j < 3; j++) {
			A[i][j] = S * (nu * A[i][j]  + M[i][j]/ dt);
		}
	}

	for (i = 0; i < 3; i++)	{
		B[i] = -g[i][0]*S/3.;
		B[i + 3] = -g[i][1]*S/3.;
	}
}

/**
 * calculate $(u1 \cdot \nabla u0, \phi)$ on an element
 */
void movedls::ele_matrix_convection (double node[3][2], double u1[6],
                                   double u0[6], double f[6])
{
	int i;
	func_on_ele u (u0, node);
	double g[3][2];
	double S = fabs (u.area)/2.;
	double sum_u = 0, sum_v = 0, u_x = 0, u_y = 0, v_x = 0, v_y = 0;
	for (i = 0; i < 3; i++) {
		g[i][0] = u.ita[i] / u.area;
		g[i][1] = -u.ata[i] / u.area;
	}
	for (i = 0; i < 3; i++)	{
		sum_u += u1[i];
		sum_v += u1[3 + i];
		u_x += u0[i] * g[i][0];
		u_y += u0[i] * g[i][1];
		v_x += u0[i + 3] * g[i][0];
		v_y += u0[i + 3] * g[i][1];
	}
	for (i = 0; i < 3; i++)	{
		f[i] = u_x * (sum_u + u1[i]) + u_y * (sum_v + u1[i + 3]);
		f[i + 3] = v_x * (sum_u + u1[i]) + v_y * (sum_v + u1[i + 3]);
	}
	for (i = 0; i < 6; i++)
		f[i] *= S / 12.;
}

/**
 * calculate $(u, \phi)$ on an element
 */
void movedls::ele_matrix_mass (double node[3][2], double u0[6], double f1[6])
{
	int i;
	func_on_ele u(u0, node);
	double S = fabs (u.area) / 2.;
	double sum_u = 0, sum_v = 0;
	for (i = 0; i < 3; i++) {
		sum_u += u0[i];
		sum_v += u0[3 + i];
	}
	for (i = 0; i < 3; i++)	{
		f1[i] = sum_u + u0[i];
		f1[i + 3] = sum_v + u0[i + 3];
	}
	for (i = 0; i < 6; i++)
		f1[i] *= S/ 12.;
}

/**
 * calculate the matrix $1/dt + nu \Delta$
 */
void movedls::GetMatrix (double dt, double nu)
{
	unsigned int i, j, k, l, m;
	double ele_node[3][2];
	double A[3][3], B[6];
	matrix_A.reinit (spM_n);
	for (i = 0; i < n_element; i++) {
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			ele_node[j][0] = node[k].coordinate[0];
			ele_node[j][1] = node[k].coordinate[1];
		}
		ele_matrix_AB(ele_node, dt, nu, A, B);
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			for (l = 0; l < 3; l++) {
				m = element[i].node[l];
				if (index[m] != n_new_node - 1) {
					matrix_A.add (index[k], n_new_node*2 + index[m], B[j]);
					matrix_A.add (n_new_node*2 + index[m], index[k], B[j]);
					matrix_A.add (n_new_node + index[k], n_new_node*2 + index[m], B[j+3]);
					matrix_A.add (n_new_node*2 + index[m], n_new_node + index[k], B[j+3]);
				}
				matrix_A.add (index[k], index[m], A[j][l]);
				matrix_A.add (n_new_node + index[k], n_new_node + index[m], A[j][l]);
			}
		}
	}
}

/**
 * calculate $(u1 \cdot \nabla u0, \phi)$
 */
void movedls::RightHand_convection (const Vector<double>& u1, const Vector<double>& u0,
	 Vector<double>& convection)
{
	unsigned int i, j, k;
	double ele_u1[6], ele_u0[6], ele_node[3][2], f[6];
	convection.reinit (n_new_node * 3 - 1);
	for (i = 0; i < n_element; i++) {
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			ele_u1[j] = u1(k);
			ele_u1[j + 3] = u1(k + n_node);
			ele_u0[j] = u0(k);
			ele_u0[j + 3] = u0(k + n_node);
			ele_node[j][0] = node[k].coordinate[0];
			ele_node[j][1] = node[k].coordinate[1];
		}
		ele_matrix_convection (ele_node, ele_u1, ele_u0, f);
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			convection (index[k]) += f[j];
			convection (n_new_node + index[k]) += f[j + 3];
		}
	}
}

/**
 * calculate $(u, \phi)$
 */
void movedls::RightHand_mass (const Vector<double>& u, Vector<double>& mass)
{
	unsigned int i, j, k;
	double ele_u[6], ele_node[3][2], f1[6];
	mass.reinit (n_new_node * 3 - 1);
	for (i = 0; i < n_element; i++) {
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			ele_u[j] = u(k);
			ele_u[j + 3] = u(k + n_node);
			ele_node[j][0] = node[k].coordinate[0];
			ele_node[j][1] = node[k].coordinate[1];
		}
		ele_matrix_mass (ele_node, ele_u, f1);
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			mass (index[k]) += f1[j];
			mass (n_new_node + index[k]) += f1[j + 3];
		}
	}
}

void movedls::StepForward(const Vector<double>& u_f)
{
	unsigned int i;
	Vector<double> new_solution(3*n_new_node - 1);
	for (i = 0; i < n_node; i++) {
		new_solution(index[i]) = solution(i);
		new_solution(index[i] + n_new_node) = solution(i + n_node);
	}
#ifdef ILU_BICG_SOLVER
	SparseILU<double> ilu (spM_n);
	ilu.decompose (matrix_A);
	SolverControl solver_control (400, 1e-6);
	PrimitiveVectorMemory<> vector_memory;
	SolverBicgstab<> bicgstab (solver_control, vector_memory);
	bicgstab.solve (matrix_A, new_solution, u_f,
		PreconditionUseMatrix<SparseILU<double>,Vector<double> >
		(ilu, &SparseILU<double>::template apply_decomposition<double>));
#endif // ILU_BICG_SOLVER
#ifdef VANKA_GMRES_SOLVER
	SolverControl solver_control (400, 1e-6);
	PrimitiveVectorMemory<> vector_memory;
	std::vector<bool> selected_dofs(3*n_new_node - 1, true);
	std::fill_n(selected_dofs.begin(), 2*n_new_node, false);
	SparseVanka<double> vanka(matrix_A, selected_dofs);
	SolverBicgstab<> gmres(solver_control, vector_memory);
	gmres.solve(matrix_A, new_solution, u_f, vanka);
#endif // VANKA_GMRES_SOLVER
	for (i = 0; i < n_node; i++) {
		solution(i) = new_solution(index[i]);
		solution(i + n_node) = new_solution(index[i] + n_new_node);
	}
}

void movedls::run ()
{
	solution.reinit(n_node*3);
	old_solution.reinit(n_node*3);
	Vector<double> convection(n_new_node * 3 - 1);
	Vector<double> mass(n_new_node * 3 - 1);
	Vector<double> rightHand(n_new_node * 3 - 1);

	double step_error = 1.;
	t = 0.0;
	step = 0.0025;
	double output_time=0;
	int i, j;
	char buffer[256];

	InitialMesh ();
    
	while (step_error > 1e-8)
	{
		step = 1.0;
		double area, chosen_step;
		for (i = 0; i< n_element; i++) {
			area = (node[element[i].node[1]].coordinate[0] - node[element[i].node[0]].coordinate[0])
				* (node[element[i].node[2]].coordinate[1] - node[element[i].node[0]].coordinate[1])
				- (node[element[i].node[1]].coordinate[1] - node[element[i].node[0]].coordinate[1])
				* (node[element[i].node[2]].coordinate[0] - node[element[i].node[0]].coordinate[0]);
			for(j = 0; j < 3; j++) {
				chosen_step = area/( solution(element[i].node[j])
					 *(node[element[i].node[(j+2)%3]].coordinate[1] - node[element[i].node[(j+1)%3]].coordinate[1]) 
					 - solution(element[i].node[j]+ n_node)
					 *(node[element[i].node[(j+2)%3]].coordinate[0] - node[element[i].node[(j+1)%3]].coordinate[0]));
				if ((chosen_step > 0) && (step > chosen_step))
					step = chosen_step;
			}
		}
		step *= 0.5;
		if(t>=output_time) {
			sprintf(buffer, "velocity_%6.5f.dat", t);
			OutputSolution (DATA_PATH + buffer);
			sprintf(buffer, "vorticity_%6.5f.dat", t);
			Vorticity (DATA_PATH + buffer);
			sprintf(buffer, "monitor_%6.5f.dat", t);
			OutputMonitor(DATA_PATH + buffer);
			output_time +=0.05;
		}
		cout<<"error:"<<L2_norm(solution)<<endl;
		// back up the old solution at first
		old_solution = solution;
		//Three step Runger-Kutta
		for (i = 3; i > 0; i --) {
			GetMatrix (step/i, Nu);
			RightHand_convection(solution, solution, convection);
			RightHand_mass(old_solution, mass);
			rightHand.equ (1./(step/i), mass, -1., convection);
			StepForward (rightHand);
		}
		MoveMesh ();

		t = t + step;
		cout << "Time = " << t << "\t\tStep = " << step << "\n";
		if (t > 2.01+step) break;
	}
}

func_on_ele::func_on_ele(double v[6],double q[3][2])
{
	int i;
	for(i=0;i<6;i++)
		coefficient[i]=v[i];
	ata[0]=q[1][0]-q[2][0]; ata[1]=q[2][0]-q[0][0]; ata[2]=q[0][0]-q[1][0];
	ita[0]=q[1][1]-q[2][1]; ita[1]=q[2][1]-q[0][1]; ita[2]=q[0][1]-q[1][1];
	omiga[0]=q[1][0]*q[2][1]-q[2][0]*q[1][1];
	omiga[1]=q[2][0]*q[0][1]-q[0][0]*q[2][1];
	omiga[2]=q[0][0]*q[1][1]-q[1][0]*q[0][1];
	area=omiga[0]+omiga[1]+omiga[2];
}

double movedls::L2_norm(Vector<double>& solution)
{
	unsigned int i, j, k;
	double l2_norm= 0., area, x[3], y[3];
	for (i = 0; i < n_element; i ++) {
		for(j=0; j<3; j++){
			x[j]= node[element[i].node[j]].coordinate[0];
			y[j]= node[element[i].node[j]].coordinate[1];
		}
		area= fabs((x[1]-x[0])*(y[2]-y[0])-(x[2]-x[0])*(y[1]-y[0]));
		for(j=0; j<3; j++) {
			for(k=j; k<3; k++) {
				  l2_norm += solution(element[i].node[j])*solution(element[i].node[k])*area/12.;
				  l2_norm += solution(n_node+ element[i].node[j])
					   *solution(n_node+ element[i].node[k])*area/12.;
			}
		}
	}
	l2_norm= sqrt(l2_norm);
	return l2_norm;  
}

void movedls::OutputSolution(const string& filename)
{
	unsigned int i;
	ofstream is(filename.c_str());
	is << "VARIABLES=\"X\",\"Y\",\"U\",\"V\"\n ";
	for(i=0;i<n_node;i++)	{
		if(node[i].coordinate[0]<0)
			 is << node[i].coordinate[0]+1 <<"  ";
		else if(node[i].coordinate[0]>1)
				is << node[i].coordinate[0]-1 <<"  ";
			 else
				is << node[i].coordinate[0] <<"  ";
		if(node[i].coordinate[1]<0)
			 is << node[i].coordinate[1]+1 <<"  ";
		else if(node[i].coordinate[1]>1)
				is << node[i].coordinate[1]-1 <<"  ";
			 else
				is << node[i].coordinate[1] <<"  ";
		is  << solution(i)           << "   "
			<< solution(i+n_node)    << "\n";
	}
	is.close();
}

void movedls::Vorticity(const string& filename)
{
	unsigned int i, j;
	double u_x, u_y, v_x, v_y, x[3], y[3], u[6];
	std::vector < double >area (n_element);
	std::vector < double >mass_accumulation (n_new_node, 0);
	std::vector < double >vorticity(n_element);
	std::vector < double >node_vorticity(n_new_node, 0);
	for (i = 0; i < n_element; i++)	{
		for(j=0;j<3;j++){
			x[j]= node[element[i].node[j]].coordinate[0];
			y[j]= node[element[i].node[j]].coordinate[1];
			u[j]= solution(element[i].node[j]);
			u[j+3]= solution(n_node+element[i].node[j]);
		}
		area[i]= (x[1]-x[0])*(y[2]-y[0])-(x[2]-x[0])*(y[1]-y[0]);
		u_x= ((u[1]-u[0])*(y[2]-y[0])-(u[2]-u[0])*(y[1]-y[0]))/area[i];
		u_y= -((u[1]-u[0])*(x[2]-x[0])-(u[2]-u[0])*(x[1]-x[0]))/area[i];
		v_x= ((u[4]-u[3])*(y[2]-y[0])-(u[5]-u[3])*(y[1]-y[0]))/area[i];
		v_y= -((u[4]-u[3])*(x[2]-x[0])-(u[5]-u[3])*(x[1]-x[0]))/area[i];
		vorticity[i] = v_x- u_y;
		mass_accumulation[index[element[i].node[0]]] += area[i];
		mass_accumulation[index[element[i].node[1]]] += area[i];
		mass_accumulation[index[element[i].node[2]]] += area[i];
	}
	for (i=0; i<n_element; i++) {
		node_vorticity[index[element[i].node[0]]] += vorticity[i] * area[i];
		node_vorticity[index[element[i].node[1]]] += vorticity[i] * area[i];
		node_vorticity[index[element[i].node[2]]] += vorticity[i] * area[i];
	}
	for(i=0; i<n_new_node; i++)
		node_vorticity[i] = node_vorticity[i]/mass_accumulation[i];

	ofstream is(filename.c_str());
	is << "VARIABLES=\"X\",\"Y\",\"V\"\n ";
	for(i = 0; i < n_node; i++){
		if(node[i].coordinate[0]<0)
			 is << node[i].coordinate[0]+1 <<"  ";
		else if(node[i].coordinate[0]>1)
				is << node[i].coordinate[0]-1 <<"  ";
			 else
				is << node[i].coordinate[0] <<"  ";
		if(node[i].coordinate[1]<0)
			 is << node[i].coordinate[1]+1 <<"  ";
		else if(node[i].coordinate[1]>1)
				is << node[i].coordinate[1]-1 <<"  ";
			 else
				is << node[i].coordinate[1] <<"  ";
		is << node_vorticity[index[i]]<<"\n";
	}
	for(i=0; i<n_face; i++) {
		if((node[face[i].node[0]].coordinate[0]+node[face[i].node[1]].coordinate[0])/2.<0)
			is << (node[face[i].node[0]].coordinate[0]+node[face[i].node[1]].coordinate[0])/2.+1 <<"   ";
		else if( (node[face[i].node[0]].coordinate[0]+node[face[i].node[1]].coordinate[0])/2.>1)
			is << (node[face[i].node[0]].coordinate[0]+node[face[i].node[1]].coordinate[0])/2.-1 <<"   ";
			else
			is << (node[face[i].node[0]].coordinate[0]+node[face[i].node[1]].coordinate[0])/2. <<"   ";
		if( (node[face[i].node[0]].coordinate[1]+node[face[i].node[1]].coordinate[1])/2.<0)
			is << (node[face[i].node[0]].coordinate[1]+node[face[i].node[1]].coordinate[1])/2.+1 <<"   ";
		else if( (node[face[i].node[0]].coordinate[1]+node[face[i].node[1]].coordinate[1])/2.>1)
			is << (node[face[i].node[0]].coordinate[1]+node[face[i].node[1]].coordinate[1])/2.-1 <<"   ";
			else
			is <<(node[face[i].node[0]].coordinate[1]+node[face[i].node[1]].coordinate[1])/2. <<"   ";
	is << (vorticity[face[i].neighbour_element[0]]+vorticity[face[i].neighbour_element[1]])/2.<<"\n";
	}
	is << "\n";
	is.close();
}

void movedls::Divergence(const string& filename)
{
	unsigned int i, j;
	double u_x, u_y, v_x, v_y, x[3], y[3], u[6];
	std::vector < double >area (n_element);
	std::vector < double >mass_accumulation (n_new_node, 0);
	std::vector < double >vorticity(n_element);
	std::vector < double >node_vorticity(n_new_node, 0);
	for (i = 0; i < n_element; i++)	{
		for(j=0;j<3;j++){
			x[j]= node[element[i].node[j]].coordinate[0];
			y[j]= node[element[i].node[j]].coordinate[1];
			u[j]= solution(element[i].node[j]);
			u[j+3]= solution(n_node+element[i].node[j]);
		}
		area[i]= (x[1]-x[0])*(y[2]-y[0])-(x[2]-x[0])*(y[1]-y[0]);
		u_x= ((u[1]-u[0])*(y[2]-y[0])-(u[2]-u[0])*(y[1]-y[0]))/area[i];
		u_y= -((u[1]-u[0])*(x[2]-x[0])-(u[2]-u[0])*(x[1]-x[0]))/area[i];
		v_x= ((u[4]-u[3])*(y[2]-y[0])-(u[5]-u[3])*(y[1]-y[0]))/area[i];
		v_y= -((u[4]-u[3])*(x[2]-x[0])-(u[5]-u[3])*(x[1]-x[0]))/area[i];
		vorticity[i] = u_x + v_y;
		mass_accumulation[index[element[i].node[0]]] += area[i];
		mass_accumulation[index[element[i].node[1]]] += area[i];
		mass_accumulation[index[element[i].node[2]]] += area[i];
	}
	for (i=0; i<n_element; i++) {
		node_vorticity[index[element[i].node[0]]] += vorticity[i] * area[i];
		node_vorticity[index[element[i].node[1]]] += vorticity[i] * area[i];
		node_vorticity[index[element[i].node[2]]] += vorticity[i] * area[i];
	}
	for(i=0; i<n_new_node; i++)
		node_vorticity[i] = node_vorticity[i]/mass_accumulation[i];

	ofstream is(filename.c_str());
	is << "VARIABLES=\"X\",\"Y\",\"V\"\n ";
	for(i = 0; i < n_node; i++){
		if(node[i].coordinate[0]<0)
			 is << node[i].coordinate[0]+1 <<"  ";
		else if(node[i].coordinate[0]>1)
				is << node[i].coordinate[0]-1 <<"  ";
			 else
				is << node[i].coordinate[0] <<"  ";
		if(node[i].coordinate[1]<0)
			 is << node[i].coordinate[1]+1 <<"  ";
		else if(node[i].coordinate[1]>1)
				is << node[i].coordinate[1]-1 <<"  ";
			 else
				is << node[i].coordinate[1] <<"  ";
		is << fabs(node_vorticity[index[i]])<<"\n";
	}
	for(i=0; i<n_face; i++) {
		if((node[face[i].node[0]].coordinate[0]+node[face[i].node[1]].coordinate[0])/2.<0)
			is << (node[face[i].node[0]].coordinate[0]+node[face[i].node[1]].coordinate[0])/2.+1 <<"   ";
		else if( (node[face[i].node[0]].coordinate[0]+node[face[i].node[1]].coordinate[0])/2.>1)
			is << (node[face[i].node[0]].coordinate[0]+node[face[i].node[1]].coordinate[0])/2.-1 <<"   ";
			else
			is << (node[face[i].node[0]].coordinate[0]+node[face[i].node[1]].coordinate[0])/2. <<"   ";
		if( (node[face[i].node[0]].coordinate[1]+node[face[i].node[1]].coordinate[1])/2.<0)
			is << (node[face[i].node[0]].coordinate[1]+node[face[i].node[1]].coordinate[1])/2.+1 <<"   ";
		else if( (node[face[i].node[0]].coordinate[1]+node[face[i].node[1]].coordinate[1])/2.>1)
			is << (node[face[i].node[0]].coordinate[1]+node[face[i].node[1]].coordinate[1])/2.-1 <<"   ";
			else
			is <<(node[face[i].node[0]].coordinate[1]+node[face[i].node[1]].coordinate[1])/2. <<"   ";
	is << fabs((vorticity[face[i].neighbour_element[0]]+vorticity[face[i].neighbour_element[1]])/2.)<<"\n";
	}
	is << "\n";
}

void movedls::OutputMonitor(const string& filename)
{
	MonitorFunction();
	unsigned int i, j;
	double area, x[3], y[3];
	std::vector<double> h1_monitor(n_node, 0);
	std::vector<double> n_neibour_ele(n_node, 0);

	for (i = 0;i < n_element;i ++) {
		for(j=0; j<3; j++){
			  x[j]= node[element[i].node[j]].coordinate[0];
			  y[j]= node[element[i].node[j]].coordinate[1];
		}
		area= (x[1]-x[0])*(y[2]-y[0])-(x[2]-x[0])*(y[1]-y[0]);
		for(j=0; j<3; j++) {
			 h1_monitor[element[i].node[j]] += monitor_function[i]*area;
			 n_neibour_ele[element[i].node[j]]  += area;
		}
	}
	for(i=0; i<n_node; i++)
		h1_monitor[i] = h1_monitor[i]/n_neibour_ele[i];

	ofstream is(filename.c_str());
	is << "VARIABLES=\"X\",\"Y\",\"V\"\n ";
	is << "ZONE N="<<n_node<<", E="<<n_element<<", F=FEPOINT, ET=TRIANGLE\n";
	for(i = 0; i < n_node; i++){
		 is   <<node[i].coordinate[0]<<"  "<<node[i].coordinate[1]<<"  "
			 <<h1_monitor[i]<< "\n";
	}
	is << "\n";
	for(i = 0; i < n_element; i++){
		 is   << element[i].node[0]+1  << "  "
			 << element[i].node[1]+1  << "  "
			 <<element[i].node[2]+1   << "  " << "\n";
	}
	is.close();
}

void movedls::ReadSolution(const string& filename)
{
	unsigned int i;
	char buffer[256];
	ifstream is;
	is.open(filename.c_str());
	std::cout << "Reading solution data ..." << std::endl;
	is.getline(buffer, 256);
	is.getline(buffer, 256);
	for (i = 0;i < n_node;i ++) {
		is >> node[i].coordinate[0]
			 >> node[i].coordinate[1]
			 >> solution(i)
			 >> solution(i+n_node);
	}
	is.close();
}

void movedls::MonitorFunction ()
{
	int i, j, k;
	std::vector<std::vector<double> > nabla_u(2, std::vector<double>(n_element, 0));
	std::vector<std::vector<double> > nabla_v(2, std::vector<double>(n_element, 0));
	double area;
	for (i = 0;i < n_element;i ++) {
		area = (node[element[i].node[1]].coordinate[0] - node[element[i].node[0]].coordinate[0])
			* (node[element[i].node[2]].coordinate[1] - node[element[i].node[0]].coordinate[1])
			- (node[element[i].node[1]].coordinate[1] - node[element[i].node[0]].coordinate[1])
			* (node[element[i].node[2]].coordinate[0] - node[element[i].node[0]].coordinate[0]);
		nabla_u[0][i] = ((solution(element[i].node[1]) - solution(element[i].node[0]))
			* (node[element[i].node[2]].coordinate[1] - node[element[i].node[0]].coordinate[1])
			- (node[element[i].node[1]].coordinate[1] - node[element[i].node[0]].coordinate[1])
			* (solution(element[i].node[2]) - solution(element[i].node[0]))
			) / area;
		nabla_u[1][i] = ((node[element[i].node[1]].coordinate[0] - node[element[i].node[0]].coordinate[0])
			* (solution(element[i].node[2]) - solution(element[i].node[0]))
			- (solution(element[i].node[1]) - solution(element[i].node[0]))
			* (node[element[i].node[2]].coordinate[0] - node[element[i].node[0]].coordinate[0])
			) / area;
		nabla_v[0][i] = ((solution(n_node+element[i].node[1]) - solution(n_node+element[i].node[0]))
			* (node[element[i].node[2]].coordinate[1] - node[element[i].node[0]].coordinate[1])
			- (node[element[i].node[1]].coordinate[1] - node[element[i].node[0]].coordinate[1])
			* (solution(n_node+element[i].node[2]) - solution(n_node+element[i].node[0]))
			) / area;
		nabla_v[1][i] = ((node[element[i].node[1]].coordinate[0] - node[element[i].node[0]].coordinate[0])
			* (solution(n_node+element[i].node[2]) - solution(n_node+element[i].node[0]))
			- (solution(n_node+element[i].node[1]) - solution(n_node+element[i].node[0]))
			* (node[element[i].node[2]].coordinate[0] - node[element[i].node[0]].coordinate[0])
			) / area;
	}
	for (i = 0;i < n_element;i ++) {
		double estimate[] = {0, 0, 0, 0};
		for (j = 0;j < 3;j ++) {
			int face_node[2];
			face_node[0]= face[element[i].face[j]].node[0];
			face_node[1]= face[element[i].face[j]].node[1];
			double face_length= sqrt((node[face_node[0]].coordinate[0]-node[face_node[1]].coordinate[0])
				 *(node[face_node[0]].coordinate[0]-node[face_node[1]].coordinate[0])
				 +(node[face_node[0]].coordinate[1]-node[face_node[1]].coordinate[1])
				 *(node[face_node[0]].coordinate[1]-node[face_node[1]].coordinate[1]));
			estimate[0] += fabs(nabla_u[0][face[element[i].face[j]].neighbour_element[0]]
				 - nabla_u[0][face[element[i].face[j]].neighbour_element[1]])*face_length;
			estimate[1] += fabs(nabla_u[1][face[element[i].face[j]].neighbour_element[0]]
				 - nabla_u[1][face[element[i].face[j]].neighbour_element[1]])*face_length;
			estimate[2] += fabs(nabla_v[0][face[element[i].face[j]].neighbour_element[0]]
				 - nabla_v[0][face[element[i].face[j]].neighbour_element[1]])*face_length;
			estimate[3] += fabs(nabla_v[1][face[element[i].face[j]].neighbour_element[0]]
				 - nabla_v[1][face[element[i].face[j]].neighbour_element[1]])*face_length;
		}
		monitor_function[i] = estimate[0]*estimate[0]*estimate[0]*estimate[0]+estimate[1]*estimate[1]*estimate[1]*estimate[1]
			 +estimate[2]*estimate[2]*estimate[2]*estimate[2]+estimate[3]*estimate[3]*estimate[3]*estimate[3];
	}
	SmoothMonitor(12);
//------------scaling---------------------------------------------------------------------------
	double max_M, min_M;
	max_M= monitor_function[0];
	min_M= monitor_function[0];
	for( i=1; i< n_element; i++) {
		max_M = max(max_M, monitor_function[i]);
		min_M = min(min_M, monitor_function[i]);
	}
	for (i = 0;i < n_element;i ++)
		monitor_function[i] = 1./sqrt(1+ 20*(monitor_function[i]/max_M));
	cout<< "monitor max:" << 1./sqrt(1+20*min_M/max_M) << endl;
}

void movedls::UpdateSolution ()
{
	unsigned int i, j;
	Vector<double> convection(n_new_node * 3 - 1);
	Vector<double> mass(n_new_node * 3 - 1);
	Vector<double> rightHand(n_new_node * 3 - 1);
	Vector<double> v(2*n_node);
	for (i = 0;i < n_node;i ++) {
		v(i) = move_direction[i].coordinate[0];
		v(i+n_node) = move_direction[i].coordinate[1];
	}
	// backup the solution at first
	old_solution = solution;
	double virtual_step = 1.0;
	double area, chosen_step;
	for (i = 0; i< n_element; i++) {
		area = (node[element[i].node[1]].coordinate[0] - node[element[i].node[0]].coordinate[0])
			* (node[element[i].node[2]].coordinate[1] - node[element[i].node[0]].coordinate[1])
			- (node[element[i].node[1]].coordinate[1] - node[element[i].node[0]].coordinate[1])
			* (node[element[i].node[2]].coordinate[0] - node[element[i].node[0]].coordinate[0]);
	    	for(j = 0; j < 3; j++) {
			chosen_step = area/( v(element[i].node[j])
				 *(node[element[i].node[(j+2)%3]].coordinate[1] - node[element[i].node[(j+1)%3]].coordinate[1]) 
				 - v(element[i].node[j]+ n_node)
				 *(node[element[i].node[(j+2)%3]].coordinate[0] - node[element[i].node[(j+1)%3]].coordinate[0]));
			if ((chosen_step > 0) && (virtual_step > chosen_step))
				virtual_step = chosen_step;
		}
	}
	virtual_step *= 0.8;
	if (virtual_step < move_step_length) {
		std::cerr << "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<" << std::endl;
		move_step_length = virtual_step;
	}
	// three step Runge-Kutta scheme
	for (i = 3;i > 0;i --) {
		GetMatrix(1.0/i, 0.0);
		RightHand_convection(v, solution, convection);
		RightHand_mass(old_solution, mass);
		rightHand.equ(1.0/(1.0/i), mass, move_step_length, convection);
		StepForward(rightHand);
	}
}

void movedls::InitialMesh()
{
	unsigned int i;
	double scale;
	scale = 0.2;
	while (scale <= 1.){
		std::cout << "scale=" << scale << std::endl;
		InitialValue();
		for (i = 0;i < n_node*3-1;i++) {
			solution(i) *= scale;
		}
		MoveMesh();
		scale += 0.2;
	}
	int n_step;
	do {
		InitialValue();
		n_step = MoveMesh();
	} while (n_step > 0);
}


/*
 * end of file
 ***************************************************************************/
