/***************************************************************************
                          stabledls.cpp  -  description
                             -------------------
    begin                :  10 30 2002
    copyright            : (C) 2002 by 
    email                : dyn@DYN
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "stabledls.h"
#define DATA_PATH	string("/home/rli/data/")
#define PI			(4.0*atan(1.0))

movedls::movedls ()
{
}

movedls::~movedls ()
{
}

void
movedls::InitialValue ()
{
	unsigned int i, j, k, l, m;
	double x, y, ro = 30., deta = 0.05;
	Re = 1. / 10000.;
	for (i = 0; i < n_node; i++) {
		x = node[i].coordinate[0];
		y = node[i].coordinate[1];
		if (y <= 0.5)
			solution (i) = tanh (ro * (y - 0.25));
		else
			solution (i) = tanh (ro * (0.75 - y));
		solution (i + n_node) = deta * sin (2. * PI * x);

		if (i != n_node - 1)
			solution (i + 2 * n_node) = 0;
	}
}


/**
 * calculate the matrix $1/dt + nu \Delta$ and $\nabla$ on an element
 */
void
movedls::ele_matrix_AB (double node[3][2], double dt, double nu, double A[3][3], double B[6])
{
	int i, j;
	double M[3][3],	u0[6], g[3][2];
	func_on_ele u (u0, node);
	double S = fabs (u.area) / 2.;
	for (i = 0; i < 3; i++)	{
		g[i][0] = u.ita[i] / u.area;
		g[i][1] = -u.ata[i] / u.area;
	}

	for (i = 0; i < 3; i++) {
		M[i][i] = 1./6.;
		for (j = i + 1; j < 3; j++) {
			M[i][j] = M[j][i] = 1./12.;
		}
	}
	for (i = 0; i < 3; i++) {
		for (j = 0; j < 3; j++) {
			A[i][j] = g[i][0]*g[j][0] + g[i][1]*g[j][1];
		}
	}
	for (i = 0; i < 3; i++) {
		for (j = 0; j < 3; j++) {
			A[i][j] = S * (nu * A[i][j] + M[i][j] / dt);
		}
	}

	for (i = 0; i < 3; i++)	{
		B[i] = -g[i][0]*S/3.;
		B[i + 3] = -g[i][1]*S/3.;
	}
}

/**
 * calculate $(u1 \cdot \nabla u0, \phi)$ on an element
 */
void
movedls::ele_matrix_convection (double node[3][2], double u1[6], double u0[6], double f[6])
{
	int i, j;
	func_on_ele u (u0, node);
	double g[3][2];
	double S = fabs (u.area)/2.;
	double sum_u = 0, sum_v = 0, u_x = 0, u_y = 0, v_x = 0, v_y = 0;
	for (i = 0; i < 3; i++) {
		g[i][0] = u.ita[i] / u.area;
		g[i][1] = -u.ata[i] / u.area;
	}
	for (i = 0; i < 3; i++)	{
		sum_u += u1[i];
		sum_v += u1[3 + i];
		u_x += u0[i] * g[i][0];
		u_y += u0[i] * g[i][1];
		v_x += u0[i + 3] * g[i][0];
		v_y += u0[i + 3] * g[i][1];
	}
	for (i = 0; i < 3; i++)	{
		f[i] = u_x * (sum_u + u1[i]) + u_y * (sum_v + u1[i + 3]);
		f[i + 3] = v_x * (sum_u + u1[i]) + v_y * (sum_v + u1[i + 3]);
	}
	for (i = 0; i < 6; i++)
		f[i] *= S / 6.;
}

/**
 * calculate $(u, \phi)$ on an element
 */
void
movedls::ele_matrix_mass (double node[3][2], double u0[6], double f1[6])
{
	int i;
	func_on_ele u(u0, node);
	double S = fabs (u.area) / 2.;
	double sum_u = 0, sum_v = 0;
	for (i = 0; i < 3; i++) {
		sum_u += u0[i];
		sum_v += u0[3 + i];
	}
	for (i = 0; i < 3; i++)	{
		f1[i] = sum_u + u0[i];
		f1[i + 3] = sum_v + u0[i + 3];
	}
	for (i = 0; i < 6; i++)
		f1[i] *= S / 12.;
}

/**
 * calculate the matrix $1/dt + nu \Delta$
 */
void
movedls::GetMatrix (double dt, double nu)
{
	unsigned int i, j, k, l, m;
	double ele_node[3][2];
	double A[3][3], B[6];
	matrix_A.reinit (spM_n);
	for (i = 0; i < n_element; i++) {
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			ele_node[j][0] = node[k].coordinate[0];
			ele_node[j][1] = node[k].coordinate[1];
		}
		ele_matrix_AB(ele_node, dt, nu, A, B);
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			for (l = 0; l < 3; l++) {
				m = element[i].node[l];
				if (index[m] != n_new_node - 1) {
					matrix_A.add (index[k], n_new_node*2 + index[m], B[j]);
					matrix_A.add (n_new_node*2 + index[m], index[k], B[j]);
					matrix_A.add (n_new_node + index[k], n_new_node*2 + index[m], B[j+3]);
					matrix_A.add (n_new_node*2 + index[m], n_new_node + index[k], B[j+3]);
				}
				matrix_A.add (index[k], index[m], A[j][l]);
				matrix_A.add (n_new_node + index[k], n_new_node + index[m], A[j][l]);
			}
		}
	}
}

/**
 * calculate $(u1 \cdot \nabla u0, \phi)$
 */
void
movedls::RightHand_convection (const Vector<double>& u1, const Vector<double>& u0, Vector<double>& convection)
{
	unsigned int i, j, k;
	double ele_u1[6], ele_u0[6], ele_node[3][2], f[6];
	convection.reinit (n_new_node * 3 - 1);
	for (i = 0; i < n_element; i++) {
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			ele_u1[j] = u1(k);
			ele_u1[j + 3] = u1(k + n_node);
			ele_u0[j] = u0(k);
			ele_u0[j + 3] = u0(k + n_node);
			ele_node[j][0] = node[k].coordinate[0];
			ele_node[j][1] = node[k].coordinate[1];
		}
		ele_matrix_convection (ele_node, ele_u1, ele_u0, f);
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			convection (index[k]) += f[j];
			convection (n_new_node + index[k]) += f[j + 3];
		}
	}
}

/**
 * calculate $(u, \phi)$
 */
void
movedls::RightHand_mass (const Vector<double>& u, Vector<double>& mass)
{
	unsigned int i, j, k;
	double ele_u[6], ele_node[3][2], f1[6];
	mass.reinit (n_new_node * 3 - 1);
	for (i = 0; i < n_element; i++) {
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			ele_u[j] = u(k);
			ele_u[j + 3] = u(k + n_node);
			ele_node[j][0] = node[k].coordinate[0];
			ele_node[j][1] = node[k].coordinate[1];
		}
		ele_matrix_mass (ele_node, ele_u, f1);
		for (j = 0; j < 3; j++) {
			k = element[i].node[j];
			mass (index[k]) += f1[j];
			mass (n_new_node + index[k]) += f1[j + 3];
		}
	}
}

void
movedls::StepForward(const Vector<double>& u_f)
{
	unsigned int i;
	SparseILU<double> ilu (spM_n);
	ilu.decompose (matrix_A);
	SolverControl solver_control (200, 1e-6);
	PrimitiveVectorMemory<> vector_memory;
	SolverGMRES<> bicgstab (solver_control, vector_memory);
	Vector<double> new_solution(3*n_new_node - 1);
	bicgstab.solve (matrix_A, new_solution, u_f,
			PreconditionUseMatrix<SparseILU<double>,Vector<double> >
			(ilu, &SparseILU<double>::template apply_decomposition<double>));
	for (i = 0; i < n_node; i++) {
		solution(i) = new_solution(index[i]);
		solution(i + n_node) = new_solution(index[i] + n_new_node);
		if (index[i] != n_new_node-1)
			solution(i + 2*n_node) = new_solution(index[i] + 2*n_new_node);
		else
			solution(i + 2*n_node) = 0.0;
	}
}


func_on_ele::func_on_ele (double v[6], double q[3][2])
{
	int i;
	for (i = 0; i < 6; i++) coefficient[i] = v[i];
	ata[0] = q[1][0] - q[2][0];
	ata[1] = q[2][0] - q[0][0];
	ata[2] = q[0][0] - q[1][0];
	ita[0] = q[1][1] - q[2][1];
	ita[1] = q[2][1] - q[0][1];
	ita[2] = q[0][1] - q[1][1];
	omiga[0] = q[1][0] * q[2][1] - q[2][0] * q[1][1];
	omiga[1] = q[2][0] * q[0][1] - q[0][0] * q[2][1];
	omiga[2] = q[0][0] * q[1][1] - q[1][0] * q[0][1];
	area = omiga[0] + omiga[1] + omiga[2];
}

double
func_on_ele::fundamental (int i, double p[2])
{
	double value = 0.;
	value = (p[0] * ita[i] - p[1] * ata[i] + omiga[i]) / area;
	return value;
}

void
func_on_ele::fund_grad (int i, double grad[2])
{
	grad[0] = ita[i] / area;
	grad[1] = -ata[i] / area;
}

void
func_on_ele::value (double p[2], double value[2])
{
	int i;
	double fund_value[3];
	for (i = 0; i < 3; i++)
		fund_value[i] = fundamental (i, p);
	value[0] = 0.;
	value[1] = 0.;
	for (i = 0; i < 3; i++)
	{
		value[0] += coefficient[i] * fund_value[i];
		value[1] += coefficient[i + 3] * fund_value[i];
	}
}

void
func_on_ele::gradient (double p[2], double grad[4])
{
	int i;
	double g[3][2], fund_value[3];
	for (i = 0; i < 3; i++)
	{
		g[i][0] = ita[i] / area;
		g[i][1] = -ata[i] / area;
		fund_value[i] = fundamental (i, p);
	}
	for (i = 0; i < 4; i++)
		grad[i] = 0.;
	for (i = 0; i < 3; i++)
	{
		grad[0] += coefficient[i] * g[i][0];
		grad[1] += coefficient[i] * g[i][1];
		grad[2] += coefficient[i + 3] * g[i][0];
		grad[3] += coefficient[i + 3] * g[i][1];
	}
}

void
movedls::run ()
{
	solution.reinit(n_node*3);
	Vector<double> convection(n_new_node * 3 - 1);
	Vector<double> mass(n_new_node * 3 - 1);
	Vector<double> rightHand(n_new_node * 3 - 1);
	
	double step_error = 1.;
	t = 0;
	unsigned int k = 0, n = 4, temp, l0, l1, l2;
	InitialMesh ();
	InitialValue();

	int i, j;
	double area, chosen_step;
	while (step_error > 1e-8)
	{
		step = 0.0025;

		// back up the old solution at first
		old_solution = solution;
		//Three step Runger-Kutta
		for (i = 4; i > 0; i --) {
			GetMatrix (step/i, Re);
			RightHand_convection(solution, solution, convection);
			RightHand_mass(old_solution, mass);
			rightHand.equ (1./(step/i), mass, -1., convection);
			StepForward (rightHand);
		}
		MoveMesh ();

		t = t + step;
		cout << "Time = " << t << "\t\tStep = " << step << "\n";
		if (k % n == 0) {
			char buffer[256];
			temp = k / n;
			sprintf(buffer, "dyn/velocity_%d.dx", temp);
			OutputSolution (DATA_PATH + buffer);
			sprintf(buffer, "dyn/vorticity_%d.dx", temp);
			Vorticity (DATA_PATH + buffer);
		}
		k ++;
		if (t > 1.20+step) break;
	}
}

void
movedls::OutputSolution (const string & filename)
{
	unsigned int i;
	ofstream os (filename.c_str ());
	os << "object 1 class array type float rank 1 shape 2 item " 
		<< n_node << " data follows\n";
	for (i = 0; i < n_node; i++)
		os << node[i].coordinate[0] << "\t" 
			<< node[i].coordinate[1] << "\n";
	os << "\nobject 2 class array type int rank 1 shape 3 item "
		<< n_element << " data follows\n";
	for (i = 0; i < n_element; i++)
		os << element[i].node[0] << "\t"
			<< element[i].node[1] << "\t"
			<< element[i].node[2] << "\n";
	os << "attribute \"element type\" string \"triangles\"\n"
		<< "attribute \"ref\" string \"positions\"\n\n";
	
	os << "object 3 class array type float rank 1 shape 2 item "
		<< n_node << " data follows\n";
	for (i = 0;i < n_node;i ++)
		os << solution(i) << "\t"
			<< solution(n_node+i) << "\n";
	os << "attribute \"dep\" string \"positions\"\n\n";
	
	os << "object \"FEMFunction-2d\" class field\n"
		<< "component \"positions\" value 1\n"
		<< "component \"connections\" value 2\n"
		<< "component \"data\" value 3\n"
		<< "end\n";
	os.close ();
}

void
movedls::Vorticity (const string & filename)
{
	unsigned int i, j;
	double area, u_x, u_y, v_x, v_y, x[3], y[3], u[6];
	std::vector < double >vorticity;
	std::vector < double >node_vorticity;
	std::vector < unsigned int >n_neibour_ele;

	node_vorticity.resize (n_node);
	vorticity.resize (n_element);
	for (i = 0; i < n_element; i++)
	{
		for (j = 0; j < 3; j++)
		{
			x[j] = node[element[i].node[j]].coordinate[0];
			y[j] = node[element[i].node[j]].coordinate[1];
			u[j] = solution (element[i].node[j]);
			u[j + 3] = solution (n_node + element[i].node[j]);
		}
		area = (x[1] - x[0]) * (y[2] - y[0]) - (x[2] - x[0]) * (y[1] - y[0]);
		u_x = ((u[1] - u[0]) * (y[2] - y[0]) - (u[2] - u[0]) * (y[1] - y[0])) / area;
		u_y = -((u[1] - u[0]) * (x[2] - x[0]) - (u[2] - u[0]) * (x[1] - x[0])) / area;
		v_x = ((u[4] - u[3]) * (y[2] - y[0]) - (u[5] - u[3]) * (y[1] - y[0])) / area;
		v_y = -((u[4] - u[3]) * (x[2] - x[0]) - (u[5] - u[3]) * (x[1] - x[0])) / area;
		vorticity[i] = v_x - u_y;
	}

	n_neibour_ele.resize (n_node);
	for (i = 0; i < n_element; i++)
	{
		for (j = 0; j < 3; j++)
		{
			node_vorticity[element[i].node[j]] += vorticity[i];
			n_neibour_ele[element[i].node[j]] += 1;
		}
	}
	for (i = 0; i < n_node; i++)
		node_vorticity[i] = node_vorticity[i] / n_neibour_ele[i];

	ofstream os (filename.c_str ());
	os << "object 1 class array type float rank 1 shape 2 item " 
		<< n_node << " data follows\n";
	for (i = 0; i < n_node; i++)
		os << node[i].coordinate[0] << "\t" 
			<< node[i].coordinate[1] << "\n";
	os << "\nobject 2 class array type int rank 1 shape 3 item "
		<< n_element << " data follows\n";
	for (i = 0; i < n_element; i++)
		os << element[i].node[0] << "\t"
			<< element[i].node[1] << "\t"
			<< element[i].node[2] << "\n";
	os << "attribute \"element type\" string \"triangles\"\n"
		<< "attribute \"ref\" string \"positions\"\n\n";
	
	os << "object 3 class array type float rank 0 item "
		<< n_node << " data follows\n";
	for (i = 0;i < n_node;i ++)
		os << node_vorticity[i] << "\n";
	os << "attribute \"dep\" string \"positions\"\n\n";
	
	os << "object \"FEMFunction-2d\" class field\n"
		<< "component \"positions\" value 1\n"
		<< "component \"connections\" value 2\n"
		<< "component \"data\" value 3\n"
		<< "end\n";
	os.close ();
}

void
movedls::ReadSolution (const string & filename)
{
	unsigned int i;
	char buffer[256];
	ifstream is;
	is.open (filename.c_str ());
	std::cout << "Reading solution data ..." << std::endl;
	is.getline (buffer, 256);
	is.getline (buffer, 256);
	for (i = 0; i < n_node; i++)
	{
		is >> node[i].coordinate[0]
			>> node[i].coordinate[1]
			>> solution (i) >> solution (i + n_node);
	}
	is.close ();
}

//void
//movedls::MonitorFunction ()
//{
//	unsigned int i, j;
//	double area, u_x, u_y, v_x, v_y, x[3], y[3], u[6];
//	std::vector<double> vorticity(n_element);
//	for (i = 0; i < n_element; i++) {
//		for (j = 0; j < 3; j++) {
//			x[j] = node[element[i].node[j]].coordinate[0];
//			y[j] = node[element[i].node[j]].coordinate[1];
//			u[j] = solution (element[i].node[j]);
//			u[j + 3] = solution (n_node + element[i].node[j]);
//		}
//		area = (x[1] - x[0]) * (y[2] - y[0]) - (x[2] - x[0]) * (y[1] - y[0]);
//		u_x = ((u[1] - u[0]) * (y[2] - y[0]) - (u[2] - u[0]) * (y[1] - y[0])) / area;
//		u_y = -((u[1] - u[0]) * (x[2] - x[0]) - (u[2] - u[0]) * (x[1] - x[0])) / area;
//		v_x = ((u[4] - u[3]) * (y[2] - y[0]) - (u[5] - u[3]) * (y[1] - y[0])) / area;
//		v_y = -((u[4] - u[3]) * (x[2] - x[0]) - (u[5] - u[3]) * (x[1] - x[0])) / area;
//		vorticity[i] = v_x - u_y;
//	}
//	for (i = 0; i < n_element; i++) {
//		monitor_function[i] = 0;
//		for (j = 0; j < 3; j++) {
//			x[0] = node[face[element[i].face[j]].node[0]].coordinate[0];
//			x[1] = node[face[element[i].face[j]].node[1]].coordinate[0];
//			y[0] = node[face[element[i].face[j]].node[0]].coordinate[1];
//			y[1] = node[face[element[i].face[j]].node[1]].coordinate[1];
//			x[2] = vorticity[face[element[i].face[j]].neighbour_element[0]];
//			y[2] = vorticity[face[element[i].face[j]].neighbour_element[1]];
//			monitor_function[i] += (x[2] - y[2])*(x[2] - y[2])
//				*sqrt((x[0] - x[1])*(x[0] - x[1]) + (y[0] - y[1])*(y[0] - y[1]));
//		}
//	}
//	SmoothMonitor (2);
//	for (i = 0; i < n_element; i++)
//		monitor_function[i] = 1. / sqrt (1. + 100*monitor_function[i]);
//}

void
movedls::MonitorFunction ()
{
	unsigned int i, j;
	double area, u_x, u_y, v_x, v_y, x[3], y[3], u[6];
	for (i = 0; i < n_element; i++) {
		for (j = 0; j < 3; j++) {
			x[j] = node[element[i].node[j]].coordinate[0];
			y[j] = node[element[i].node[j]].coordinate[1];
			u[j] = solution (element[i].node[j]);
			u[j + 3] = solution (n_node + element[i].node[j]);
		}
		area = (x[1] - x[0]) * (y[2] - y[0]) - (x[2] - x[0]) * (y[1] - y[0]);
		u_x = ((u[1] - u[0]) * (y[2] - y[0]) - (u[2] - u[0]) * (y[1] - y[0])) / area;
		u_y = -((u[1] - u[0]) * (x[2] - x[0]) - (u[2] - u[0]) * (x[1] - x[0])) / area;
		v_x = ((u[4] - u[3]) * (y[2] - y[0]) - (u[5] - u[3]) * (y[1] - y[0])) / area;
		v_y = -((u[4] - u[3]) * (x[2] - x[0]) - (u[5] - u[3]) * (x[1] - x[0])) / area;
		monitor_function[i] = u_x*u_x + u_y*u_y + v_x*v_x + v_y*v_y;
	}
	SmoothMonitor (2);
	for (i = 0; i < n_element; i++)
		monitor_function[i] = 1. / sqrt (1. + monitor_function[i]);
}

void
movedls::UpdateSolution ()
{
	unsigned int i;
	Vector<double> convection(n_new_node * 3 - 1);
	Vector<double> mass(n_new_node * 3 - 1);
	Vector<double> rightHand(n_new_node * 3 - 1);
	Vector<double> v(2*n_node);
	for (i = 0;i < n_node;i ++) {
		v(i) = move_direction[i].coordinate[0];
		v(i+n_node) = move_direction[i].coordinate[1];
	}
	// backup the solution at first
	old_solution = solution;
	// three step Runge-Kutta scheme
	for (i = 4;i > 0;i --) {
		GetMatrix(1.0/i, 0.0);
		RightHand_convection(v, solution, convection);
		RightHand_mass(old_solution, mass);
		rightHand.equ(1.0/(1.0/i), mass, 1.0*move_step_length, convection);
		StepForward(rightHand);
	}
}

void
movedls::InitialMesh ()
{
	unsigned int i;
	double scale = 0.2;
	while (scale <= 1.)	{
		std::cout << "scale=" << scale << std::endl;
		InitialValue ();
		for (i = 0; i < n_node * 3 - 1; i++)
			solution (i) *= scale;
		MoveMesh ();
		scale += 0.2;
	}
	InitialValue ();
}


/*
 * end of file
 ***************************************************************************/
