/***************************************************************************
                          stabledls.h  -  description
                             -------------------
    begin                :  10 30 2002
    copyright            : (C) 2002 by 
    email                : dyn@DYN
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef _stabledls_h_
#define _stabledls_h_

#include <string>
#include <vector>
#include <lac/sparsity_pattern.h>
#include <lac/sparse_matrix.h>
#include <AFEPack/AMGSolver.h>

#include <iostream>
#include <fstream>
#include <cmath>

#include <lac/vector.h>
#include <lac/precondition.h>
#include <lac/solver_control.h>
#include <lac/solver_cg.h>
#include <lac/solver_gmres.h>
#include <lac/solver_richardson.h>
#include <lac/solver_bicgstab.h>
#include <lac/vector_memory.h>
#include <lac/sparse_ilu.h>

#include "MovingMesh.h"

using namespace std;

//------solve double shear-layer problem using finite element------------------

class movedls:public LRMovingMesh
{
      public:
	double t;
	double step;
	double Re;		//  1/Re
	SparseMatrix < double >matrix_A;
	SparseMatrix < double >matrix_V;
	Vector < double > solution;
	Vector < double > old_solution;

	movedls ();
	virtual ~movedls ();

	void GetMatrix (double dt, double nu);
	void ele_matrix_AB (double node[3][2], double dt, double nu, double A[3][3], double B[6]);
	void ele_matrix_convection(double node[3][2], double u1[6], double u0[6], double f[6]);
	void ele_matrix_mass(double node[3][2], double u0[6], double f1[6]);
	void RightHand_convection(const Vector<double>& u1, const Vector<double>& u0, Vector<double>& convection);
	void RightHand_mass(const Vector<double>& u, Vector<double>& mass);
	void InitialValue ();
	void StepForward (const Vector < double >&u_f);
	void run ();
	void OutputSolution (const string & filename);
	void Vorticity (const string & filename);
	void ReadSolution (const string & filename);
	void InitialMesh ();
	virtual void MonitorFunction ();
	virtual void UpdateSolution ();

};

class func_on_ele
{
      public:
	double coefficient[6];
	double ata[3];		//relative coordinate x
	double ita[3];		//relative coordinate y
	double omiga[3];
	double area;		//double area with direction

	  func_on_ele (double v[9], double q[3][2]);
	double fundamental (int i, double p[2]);
	void fund_grad (int i, double grad[2]);
	void value (double p[2], double value[2]);
	void gradient (double p[2], double grad[4]);
};

#endif
